#! /usr/bin/env python
"""Twisted-based channel server"""
from twisted.web.resource import Resource
from twisted.internet import defer
from torchannels import basechannels,channels
from gettext import gettext as _
import simplejson, time
from twisted.web.server import NOT_DONE_YET
from twisted.internet import reactor

def arg_list( request, key, default ):
    if key in request.args:
        value = request.args[key]
        if not isinstance( value, list ):
            return [ value ]
        return value
    return default

def arg_one( request, key, default ):
    if key in request.args:
        value = request.args[key]
        if isinstance( value, list ):
            return value[-1]
        return value
    return default

class TwistChannel( channels.Channel ):
    def send( self, message ):
        """Arrange for send to happen in main twisted thread"""
        reactor.callFromThread(
            super(TwistChannel,self).send,
            message
        )

class Channels( basechannels.BaseChannels, Resource ):
    """The channel-serving resource"""
    channel_class = TwistChannel
    isLeaf = True
    def render_POST( self, request ):
        """Render the POST request as a JSON message"""
        request.setHeader('Content-Type','text/plain')
        subscriptions = arg_list( request, 'channels',[] )
        if not subscriptions:
            return simplejson.dumps( {
                'error': _('No subscriptions specified'),
                'ts': time.time(),
            } )
        timestamp = float(arg_one(request,'cursor', 0.0))
        # check that we got valid channel names
        for sub in subscriptions:
            if not self.can_access( sub ):
                return simplejson.dumps({
                    'error': _('Invalid channel name %(sub)r')%dict(sub=sub),
                })
        stream = bool(arg_one( request,'stream', False ))
        quickPoll = bool(arg_one( request,'quickPoll',False))
        pending = self.updates( subscriptions, timestamp )
        if pending:
            request.write(simplejson.dumps( dict(
                messages = pending,
                ts = time.time(),
            ) ))
            if not stream:
                request.finish()
                return
            else:
                request.write( self.MESSAGE_SPLIT_TEXT )
        else:
            if quickPoll:
                request.write( { 'ok': True, 'messages': [] } )
                return request.finish()
        waiter = _MessageWaiter(request, stream)
        self.subscribe( subscriptions, waiter )
        # if the caller hangs up, remove waiter...
        request.notifyFinish().addErrback(waiter.clientDisconnected)
        return NOT_DONE_YET
    render_GET = render_POST

class _MessageWaiter( object ):
    invalid = False
    def __init__( self, request, stream=True ):
        self.request = request
        self.stream = stream
        self.finishedDF = defer.Deferred()
    def clientDisconnected( self, err ):
        self.invalid = True
        self.finishedDF.errback( err )
    def __call__( self, message ):
        try:
            self.request.write( simplejson.dumps({
                'messages': [message],
                'ok':True,
            }) )
            if self.stream:
                self.request.write(Channels.MESSAGE_SPLIT_TEXT)
        except Exception, err:
            self.invalid = True
            try:
                self.request.finish()
            except RuntimeError, err:
                pass
            else:
                self.finishedDF.errback( err )
        else:
            if not self.stream:
                try:
                    self.request.finish()
                except RuntimeError, err:
                    pass
                else:
                    self.finishedDF.callback( message )
                self.invalid = True

def main(port=8080):
    # TODO: there would need to be something else running in-process
    # to actually make this useful...
    from twisted.web.server import Site
    from twisted.internet import reactor, ssl
    resource = Channels()

    factory = Site(resource)
    reactor.listenTCP( int(port), factory )
    reactor.run()

if __name__ == "__main__":
    import sys
    main( *sys.argv[1:] )
